#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <libgen.h>
#include <math.h>

#include "ucb_A3d.h"

#define OUTPUT_FNAME "model-samples.out"

#define MIN_RADIUS 3480.0
#define MAX_RADIUS 6341.0
#define MAX_RADIUS_WARN 6321.0

const char version[] = "Version 0.2 (2015/09/03)";

const char default_output_file[] = OUTPUT_FNAME;

void usage(const char *name) {
	printf("\nUsage: %s -r radius [-d discretization | -l lon,lat] [-o output_file]\n\n", name);
	printf("\tAll argument units are km and degrees\n");
	printf("\tDefault output file: %s\n", default_output_file);
	printf("\n");
}

void error_exit(const char *name, const char *reason_fmt, ...)
{
	if (reason_fmt != NULL) {
		va_list args;
		va_start(args, reason_fmt);
		vprintf(reason_fmt, args);
		va_end(args);
	}
	usage(name);
	exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
	int rflag, dflag, lflag;
	double r, dsc, lon, lat, r_moho, r_m, rho, vpv, vsv, qk, qmu, vph, vsh, eta,
	       dvs, dxi, vs, xi, vsv_1, vsh_1, vs_1, xi_1;
	char ch, *name;
	const char *outfile;
	FILE *f_out;

	printf("\nUCB A3d model evaluation tool %s\n\n", version);

	name = basename(argv[0]);

	outfile = default_output_file;

	rflag = 0;
	dflag = 0;
	lflag = 0;
	while ((ch = getopt(argc, argv, "r:d:l:o:")) != -1) {
		switch (ch) {
			case 'r':
				rflag = 1;
				if (sscanf(optarg, "%lf", &r) != 1)
					error_exit(name, "Error: nonsense radius argument \"%s\"\n", optarg);
				break;
			case 'd':
				dflag = 1;
				if (sscanf(optarg, "%lf", &dsc) != 1)
					error_exit(name, "Error: nonsense discretization argument \"%s\"\n", optarg);
				break;
			case 'l':
				lflag = 1;
				if (sscanf(optarg, "%lf,%lf", &lon, &lat) != 2)
					error_exit(name, "Error: nonsense location argument \"%s\"\n", optarg);
				break;
			case 'o':
				outfile = optarg;
				break;
			case '?':
			default:
				error_exit(name, NULL);
		}
	}

	/* check radius set */
	if (!rflag)
		error_exit(name, "Error: no radius supplied\n");

	/* check radius value */
	if (r < MIN_RADIUS)
		error_exit(name, "Error: model is undefined below %.1f km\n", MIN_RADIUS);
	if (r > MAX_RADIUS)
		error_exit(name, "Error: model is undefined above %.1f km\n", MAX_RADIUS);
	if (r > MAX_RADIUS_WARN)
		printf("Warning: structure above ~ %.1fkm is less robust\n\n", MAX_RADIUS_WARN);

	/* check lateral sampling set */
	if (dflag || lflag) {
		if (dflag && lflag)
			error_exit(name, "Error: discretization and location flags set, but only one needed\n");
	} else
		error_exit(name, "Error: neither discretization nor location flags set\n");

	/* initialize model data */
	init_ucb_ref_();
	init_ucb_a3d_();
	init_moho_();

	/* km -> m for get_ucb_ref_() */
	r_m = 1000 * r;

	/* get reference structure */
	get_ucb_ref_(&r_m, &rho, &vpv, &vsv, &qk, &qmu, &vph, &vsh, &eta);

	/* m -> km */
	vsv /= 1000.0;
	vsh /= 1000.0;

	/* (vsv, vsh) -> voigt, xi */
	vs = sqrt((2.0 * vsv * vsv + vsh * vsh) / 3.0);
	xi = vsh * vsh / (vsv * vsv);

	if ((f_out = fopen(outfile, "w")) == NULL)
		error_exit(name, "Error: cannot open output file \"%s\" - %s\n", outfile, strerror(errno));

	/* sample the model ... */

	if (lflag) {
		/* evaluate model at a single location */
		get_moho_radius_(&lat, &lon, &r_moho);
		if (r_moho <= r) {
			/* in crust */
			fprintf(f_out, "%5.1f %8.3f %8.3f    nan    nan    nan    nan\n", r, lon, lat);
		} else {
			get_a3d_perturbation_(&lat, &lon, &r, "S", &dvs);
			get_a3d_perturbation_(&lat, &lon, &r, "X", &dxi);
			vs_1 = (1.0 + dvs) * vs;
			xi_1 = (1.0 + dxi) * xi;
			vsv_1 = sqrt(3.0 / (2.0 + xi_1)) * vs_1;
			vsh_1 = sqrt(xi_1) * vsv_1;
			fprintf(f_out, "%5.1f %8.3f %8.3f %6.2f %6.2f %6.3f %6.3f\n", r, lon, lat, 100 * dvs, 100 * dxi, vsv_1, vsh_1);
		}
	} else {
		/* evaluate model on a regular lat / lon grid */
		lat = -90.0;
		while (lat <= 90.0) {
			lon = -180.0;
			while (lon <= +180.0) {
				get_moho_radius_(&lat, &lon, &r_moho);
				if (r_moho <= r) {
					/* in crust */
					fprintf(f_out, "%5.1f %8.3f %8.3f    nan    nan    nan    nan\n", r, lon, lat);
				} else {
					get_a3d_perturbation_(&lat, &lon, &r, "S", &dvs);
					get_a3d_perturbation_(&lat, &lon, &r, "X", &dxi);
					vs_1 = (1.0 + dvs) * vs;
					xi_1 = (1.0 + dxi) * xi;
					vsv_1 = sqrt(3.0 / (2.0 + xi_1)) * vs_1;
					vsh_1 = sqrt(xi_1) * vsv_1;
					fprintf(f_out, "%5.1f %8.3f %8.3f %6.2f %6.2f %6.3f %6.3f\n", r, lon, lat, 100 * dvs, 100 * dxi, vsv_1, vsh_1);
				}
				lon += dsc;
			}
			lat += dsc;
		}
	}

	printf("Wrote model samples to %s\n\n", outfile);

	fclose(f_out);

	return 0;
}
